//=============================================================================
//                  m e l ( i n a )  m e s h   u t i l i t i e s
//=============================================================================

#include "mesh.h"
#include "utils.h"
#include <list>

extern map<uint,element_type> element_type_list;
extern map<string,mesh_type> mesh_type_list;
extern bool debug;
extern ofstream ofdebug;

// ============================================================================
//                             format MEL (melina)
// ============================================================================
// * ligne de commentaire
// ----------------------------------------------------------------------------
// TITRE  nbligne
// .....
// ----------------------------------------------------------------------------
// [ FORMAT [DE] LECTURE
//     [ [DES] COORDONNEES < ||C*50 ^ '*' > ]
//     [ [DE LA] NUMEROTATION [GLOBALE] < ||C*50 ^ '*' > ]
//     [ < SANS ^ AVEC > COMMENTAIRE ]
//  ]
// ---------------------------- valeurs par dfaut ----------------------------
//             FORMAT DE LECTURE DES COORDONNEES '6E12.4'
//             DE LA NUMEROTATION GLOBALE '18I4'
//             AVEC COMMENTAIRE
// ----------------------------------------------------------------------------
// DESCRIPTION GLOBALE [DU] [MAILLAGE]
//     [ [NOM] [DES] VARIABLES [D''] ESPACE : + ||C ]
//     NOMBRE [D''] ELEMENTS : ||I ]
// --------------------------------- exemples ---------------------------------
//             DESCRIPTION GLOBALE DU MAILLAGE
//             NOMS DES VARIABLES D''ESPACE : 'X' 'Y' 'Z'
//             NOMBRE D''ELEMENTS : 100
// ----------------------------------------------------------------------------
// + BLOC [DE]
//    < < HEXAEDRES ^ PRISMES ^ TETRADEDRES ^ QUADRANGLES ^ TRIANGLES ^ SEGMENTS >
//        DE LAGRANGE [ AUX POINTS DE GAUSS-LOBATTO ]
//        < < P1 ^ Q1 ^ P2 ^ Q2 ^ P3 ^ Q3 ^ Q4 ... Q9 ^ Q10 >
//        ^ < D'ORDRE ^ DE DEGRE > ||I
//        >
//    ^ [TYPE] [GEOMETRIQUE] ||A
//    > : ||IE ELEMENTS
// --------------------------------- exemples ---------------------------------
// BLOC DE	 TRIANGLES DE LAGRANGE P2 : 10 ELEMENTS       BLOC DE TYPE GEOMETRIQUE TR02 : 10 ELEMENTS
//  	   QUADRANGLES DE LAGRANGE Q2 :  5 ELEMENTS   ou                           QU02 :  5 ELEMENTS
//  	     TRIANGLES DE LAGRANGE P2 : 15 ELEMENTS                                TR02 : 15 ELEMENTS
// ----------------------------------------------------------------------------
// X1 Y1 Z1 X2 Y2 Z2 X3 Y3 Z3 .... (coordonnes des noeuds de l'lment 1)
// I1 I2 I3 ....                   (numro des noeuds de l'lment 1)
// ...
// ---------------------------------- exemple----------------------------------
//     BLOC DE TRIANGLES DE LAGRANGE P1 : 3 ELEMENTS
//     1.5000 0.0000 1.3858 0.5740 1.0000 0.0000
//     5  6  1
//     0.8660 0.5000 1.0000 0.0000 1.3858 0.5740
//     2  1  6
//     1.3858 0.5740 1.0607 1.0607 0.8660 0.5000
//     6  7  2
// ----------------------------------------------------------------------------
// DOMAINE 'nomdomaine'
// + E[LEMENT] ||I [ / ||J ]
// + E[LEMENT] ||I1
//    < F[ACE] ||I2 ^ A[RETE] ||I2 ^ P[OINT] ||I2 >
// --------------------------------- exemples ---------------------------------
//     DOMAINE 'OMEGA'
//     E 1 E 3 E 5 E 7 E 8 E 9     quivalent  E 1 3 5 7 ELEMENT 8 9
//     DOMAINE 'OMEGA' ELEMENTS 1 / 7  equivalent  E 1 2 3 4 5 6 7
//     DOMAINE 'GAMMA'
//     ELEMENT 1 FACE 3 ELEMENT 2 FACE 2 ELEMENT 7 FACE 2
//     quivalent                E 1 F 3 E 2 F 2 E 7 F 2
// ----------------------------------------------------------------------------
// FIN
// 
// attention : ne pas dpasser 80 caractres par ligne ....
// ============================================================================

// save mesh to mel format file
void mesh::saveto_mel(const char * filename, bool silent) {
    push_context("mesh::saveto_mel");
    string savename;

	for (cit_le it=elements_list.begin(); it!=elements_list.end(); ++it) {
        if (element_type_list[it->second.type].mel_name=="UNDEF") { // forbidden elements
                message_event("err_melelem",msg_data(element_type_list[it->second.type].name));
        }
    }
    
    if (filename!=NULL) {
        savename=string(filename);
    }

    if (!valid_savefile(savename,string("mel"))) {
        return;
    }

    ofstream f(savename.c_str(),ios::out | ios::binary);

    if (f.fail()) {
        message_event("open_fail",msg_data(savename));
    }
    
    // coordinates name
    if (dim==0) {
        message_event("meshdim_null");
    }

    string coor;
    vector<uint> ncor;
    string sdim[4];
    sdim[0]="0D (point)";
    sdim[1]="1D";
    sdim[2]="2D";
    sdim[3]="3D";

    if (dim==3) {
        coor="'X' 'Y' 'Z'";
        ncor.push_back(1);
        ncor.push_back(2);
        ncor.push_back(3);
    }

    if (dim==2) {
        if (!xv) {
            coor="'Y' 'Z'";
            ncor.push_back(2);
            ncor.push_back(3);
        }

        if (!yv) {
            coor="'X' 'Z'";
            ncor.push_back(1);
            ncor.push_back(3);
        }

        if (!zv) {
            coor="'X' 'Y'";
            ncor.push_back(1);
            ncor.push_back(2);
        }
    }

    if (dim==1) {
        if (!xv && !yv) {
            coor="'Z'";
            ncor.push_back(3);
        }

        if (!yv && !zv) {
            coor="'X'";
            ncor.push_back(1);
        }

        if (!xv && !zv) {
            coor="'Y'";
            ncor.push_back(2);
        }
    }
    
    // title
    vector<string> tl;
    string ti=title;

    while (ti.size()>0) {
        tl.push_back(ti.substr(0,80));
        ti.erase(0,80);
    }
    
    // list of all physical domains : name, domain, keep
    map<string,dom_info> dlist;
    map<string,dom_info>::iterator itd,itd1,itd2;
    for (it_ld it=domains_list.begin(); it!=domains_list.end(); ++it) {
        dlist[it->second.name]=dom_info(&(it->second),true,it->second.name);
    }

    if (!silent) {
        // language dependances
        string LID,DOM,ELT,BOU,OF;

        switch (msg::lang) {
            case english :
                LID="List of the domains : ";
                break;
            default :
                LID="Liste des domaines : ";
                break;
        }

        cout << LID;
        
        // domains to save, make a list of domains, rename and redefine subdomain (extract part of a subdomain)
        //     C(ontinue) ; Q(uit) ; L(ist) ; I(gnore) domain_name ; K(eep) domain_name;
        //     R(ename) domain_name domain_name ; E(xtract) new_name sub_domain_name domain_name
        //     D(isjoint) sub_domain_name

        // the command Extract is used to choose the parent of a subdomain; it is possible to "create" new one
        //  exemple 1 :  "Gamma" interface between "Omega1" and "Omega2"
        //               E(xtract) "Gamma1" "Gamma" "Omega1"    ->  define subdomain "Gamma1" as an Omega1 boundary
        //               E(xtract) "Gamma2" "Gamma" "Omega2"    ->  define subdomain "Gamma2" as an Omega2 boundary
        //    rmk :in that case, if no redefinition is given the first domain parent is used in Melina subdomain definition
        //  exemple 2 : "Gamma" boundary of "Omega1" and "Omega2" (not an interface)
        //               E(xtract) "Gammai" "Gamma" "Omegai"    ->  define sub domain "Gammai" by extracting element on the "Omegai" boundary
        //    if no redefinition, "Gamma" remains the same
        // When extracting subdomain, original subdomain is ignored by default. You can keep it by using command K(eep).

        // The command Disjoint can be applied to any domain F of dim-1, any node of F is split in two nodes,
        // the domains having F as "boundary" are updated and new subdomain(s) is (are) created
        // there are two typical situations :
        //          a dim-1 domain F in a dim domain O :  Disjoint F -> F_O subdomain is created and the domain O is modified
        //          a dim-1 domain F separating two dim domains O1 and O2 :  Disjoint F ->  F_O1, F_O2 are created and O1,O2 are modified
        // rmk : the end points of F are split

        string rep="";
        string choix="";
        bool wait=true;

        while (wait) {
            for (itd=dlist.begin(); itd!=dlist.end(); ++itd) {
                cout << itd->second;
            }
            
            cout << endl;

            message_event("que_domains");
            cin >> choix;

            if (choix[0]=='Q' || choix[0]=='q' ) {
                exit(0);
            }

            if (choix[0]=='C' || choix[0]=='c') {
                wait=false;
            }

            //if(rep[0]=='L' || rep[0]=='l')  for(itd=dlist.begin();itd!=dlist.end();++itd) cout<<itd->second;
            if (choix[0]=='I' || choix[0]=='i') { // ignore a domain
            	message_event("choice_i");
                cin >> rep; // domain _name
                itd=dlist.find(rep);

                if (itd==dlist.end()) {
                    message_event("que_domains2",msg_data(rep));

                } else {
                    if (itd->second.new_def) {
                        dlist.erase(itd); // erase new definition

                    } else {
                        itd->second.keep=false;
                    }
                }
            }
            if (choix[0]=='K' || choix[0]=='k' || choix[0]=='G' || choix[0]=='g') { // keep a domain
            	message_event("choice_k");
                cin >> rep; // domain _name
                itd=dlist.find(rep);

                if (itd==dlist.end()) {
                    message_event("que_domains2",msg_data(rep));
                }

                itd->second.keep=true;
            }

            if (choix[0]=='R' || choix[0]=='r') {
            	message_event("choice_r1");
                cin >> rep; // domain _name
            	
                itd=dlist.find(rep);

                if (itd==dlist.end()) {
                    message_event("que_domains2",msg_data(rep));

                } else {
                	message_event("choice_r2");
                    cin>>rep;
                    itd->second.new_name=force_quotation_mark(rep);
                    cout<<itd->second;
                }
            }

            if (choix[0]=='E' || choix[0]=='e') { // extract a subdomain
                cin >> rep; // new domain name
                string na=force_quotation_mark(rep);
                cin >> rep; // original domain
                rep=force_quotation_mark(rep); 
                itd=dlist.find(rep);
                bool valid=true;

                if (itd==dlist.end()) {
                    message_event("que_domains2",msg_data(rep));
                    valid=false;
                }

                if (valid && itd->second.domain->dim==dim) {
                    message_event("que_domains3",msg_data(rep));
                    valid=false;
                }

                if (valid && itd->second.new_def) {
                    message_event("que_domains5",msg_data(rep));
                    valid=false;
                }

                if (valid) {
                    cin >> rep;
                    rep=force_quotation_mark(rep);
                    itd1=dlist.find(rep);

                    if (itd1==dlist.end()) {
                        message_event("que_domains2",msg_data(rep));

                    } else {
                        dlist[na]=dom_info(itd->second.domain,true,na,true,itd1->second.domain);
                        itd->second.keep=false;
                    }

                } else {
                    cin >> rep;
                }
            }

            if (choix[0]=='D' || choix[0]=='d' || choix[0]=='F' || choix[0]=='f') { // transform a sub_domain into a crack
                cin >> rep; // domain _name
                string na=force_quotation_mark(rep);
                itd=dlist.find(na);
                bool valid=true;

                if (itd==dlist.end()) {
                    message_event("que_domains2",msg_data(na));
                    valid=false;
                }

                if (valid && itd->second.domain->dim==dim) {
                    message_event("que_domains3",msg_data(na));
                    valid=false;
                }

                if (valid) {
                    physical_domain * pd=itd->second.domain;
                    dlist[na].keep=false; // do not keep original subdomain
                    dlist[na].crack=true; // say it is a crack

                    for (uint i=0; i<pd->parents.size(); i++) { // create new domain definitions, sides of the crack
                        string napdi=ignore_quotation_mark(pd->name);
                        napdi+="_";
                        napdi+=ignore_quotation_mark(pd->parents[i]->name);
                        napdi=force_quotation_mark(napdi);
                        dlist[napdi]=dom_info(pd,true,napdi,true,pd->parents[i]);
                        dlist[napdi].cracked=true;
                        napdi=force_quotation_mark(pd->parents[i]->name);
                        dlist[napdi].cracked=true;
                    }

                } else {
                    cin >> rep;
                }
            }
        }
    } // end no silent mode
    
    // special treatment for a crack
    for (itd=dlist.begin(); itd!=dlist.end(); ++itd) {
        if (itd->second.crack) {
            uint ld=last_dom();
            physical_domain * pd=itd->second.domain;
            crack_sub_domain(*itd->second.domain); // crack the domain

            for (uint i=0; i<pd->parents.size(); i++) {
                string napdi=ignore_quotation_mark(pd->name);
                napdi+="_";
                napdi+=ignore_quotation_mark(pd->parents[i]->name);
                napdi=force_quotation_mark(napdi);
                ld++;
                dlist[napdi].domain=&domains_list[ld];
            }
        }
    }
    
    // napd=force_quotation_mark(napd);
    // dlist[napd].domain=&domains_list[lastdom]; // update domain info

    // write mel header
    uint lt=tl.size();
    f << "TITRE " << lt+1 << endl;

    for (uint i=0; i<lt; i++) {
        f << tl[i] << endl;
    }

    f << "(generated by convmesh from " << name << ")" << endl;
    f << "FORMAT DE LECTURE DES COORDONNEES '*'" << endl;
    f << "       DE LA NUMEROTATION GLOBALE '*'" << endl;
    f << "       SANS COMMENTAIRE" << endl;
    f << "DESCRIPTION GLOBALE DU MAILLAGE" << endl;
    f << "NOM DES VARIABLES D''ESPACE : " << coor << endl;
    
    // number of elements of all domain of dimension dim, sort elements by type
    uint ne=0;
    multimap<uint,uint> li_elt;
    vector<uint> li_type;

    for (cit_ld it=domains_list.begin(); it!=domains_list.end(); ++it) {
        const physical_domain& p=it->second;

        if (!p.multidim && p.dim==dim) {
            ne+=p.elements.size();

            for (uint i=0; i<p.elements.size(); i++) {
                element & E=elements_list(p.elements[i]);
                uint t=E.type;

                if (find(li_type.begin(),li_type.end(),t)==li_type.end()) {
                    li_type.push_back(t);
                }

                li_elt.insert(pair<uint,uint>(t,E.num));
            }
        }
    }
    
    // write element of mesh dimension by type
    f << "NOMBRE D''ELEMENTS : " << ne << endl;
    // element type list
    typedef multimap<uint,uint>::iterator itm;
    string com="BLOC DE TYPE GEOMETRIQUE ";

    for (uint i=0; i<li_type.size(); i++) {
        uint t=li_type[i];
        uint nt=li_elt.count(t);
        f << com << element_type_list[t].mel_name << " : " << nt << " ELEMENTS" << endl;
        com="                        ";
    }
    
    // coordinates and indexes of all elements of dimension dim, sorted by type
    uint no=0; // reindexing elements

    for (uint i=0; i<li_type.size(); i++) {
        pair<itm,itm> pit=li_elt.equal_range(li_type[i]);

        for (itm jt=pit.first; jt!=pit.second; ++jt) {
            element & E=elements_list(jt->second);
            no++;
            E.num_order=no;

            for (uint j=0; j<E.nodes.size(); j++) {
                uint ncar=0;

                for (uint k=0; k<ncor.size(); k++) {
                    stringstream ss (stringstream::out);
                    ss << nodes_list(E.nodes[j])(ncor[k]) << " ";
                    uint ns=ss.str().size();

                    if (ns+ncar>=80) {
                        f << endl;
                        ncar=0;
                    }

                    f << ss.str();
                    ncar+=ns;
                }
            }

            f << endl;
            uint ncar=0;

            for (uint j=0; j<E.nodes.size(); j++) {
                stringstream ss (stringstream::out);
                ss << E.nodes[j] <<" ";
                uint ns=ss.str().size();

                if (ns+ncar>=80) {
                    f << endl;
                    ncar=0;
                }

                f << ss.str();
                ncar+=ns;
            }
            f << endl;
        }
    }
        
    // list of domains (multi dimensionnal domain are not processed !)
    for (itd=dlist.begin(); itd!=dlist.end(); ++itd) {
        const physical_domain & p=*itd->second.domain;

        if (p.multidim) {
            message_event("warn_save_mel1",msg_data(p.name));
        }

        if (!p.multidim && itd->second.keep) {
            string na=force_quotation_mark(itd->second.new_name);
            f << "DOMAINE " << replace(na,'"','\'') << endl;

            if (p.dim==dim) { // mesh.dim domain
                f << "ELEMENT ";
                uint ncar=8;

                for (uint i=0; i<p.elements.size(); i++) {
                    element & E=elements_list(p.elements[i]);
                    stringstream ss (stringstream::out);
                    ss << E.num_order << " ";
                    uint ns=ss.str().size();

                    if (ns+ncar>=80) {
                        f << endl;
                        ncar=0;
                    }

                    f << ss.str();
                    ncar+=ns;
                }
				f << endl;
            }
            
            if (p.dim==dim-1) { // mesh.dim-1 domain
                string bord;

                switch (p.dim) {
                    case 2 :
                        bord="F";
                        break;
                    case 1 :
                        bord="A";
                        break;
                    case 0 :
                        bord="P";
                        break;
                    default:
                        message_event("err_dim_dom");
                }

                uint ncar=0;

                for (uint i=0; i<p.elements.size(); i++) {
                    element & E=elements_list(p.elements[i]);
                    list<pair<uint,uint> >::const_iterator itn=E.boundary_element.begin();

                    if (itd->second.new_def) { // domain filter (extraction command E)
                        while (itn!=E.boundary_element.end() && elements_list(itn->first).num_domain!=itd->second.parent_domain->num) {
                            itn++;
                        }
                    }
                    
                    if (itn!=E.boundary_element.end()) {
                        stringstream ss (stringstream::out);
                        ss << "E " << elements_list(itn->first).num_order << " " << bord << " " << itn->second << " ";
                        uint ns=ss.str().size();

                        if (ns+ncar>=80) {
                            f << endl;
                            ncar=0;
                        }

                        f << ss.str();
                        ncar+=ns;
                    }
                }
                f << endl;
            }

            if (p.dim>1 && p.dim==dim-2) { // mesh.dim-2 domain (edge in 3D or point in 2D)
                message_event("warn_save_mel2",msg_data(na)); // not processed yet
            }

            if (p.dim>2 && p.dim==dim-3) { // mesh.dim-3 domain (point in 3D)
                message_event("warn_save_mel2",msg_data(na)); // not processed yet
            }
        }
    }
    
    // end of mel file
    f << "FIN" << endl;
    f.close();
    pop_context();
}

// mel mesh loading
void mesh::loadfrom_mel(const char * filename) {
    push_context("mesh::loadfrom_mel");
    // meaningless words in mel format
    list<string> m_words;
    m_words.push_back("DE");
    m_words.push_back("DES");
    m_words.push_back("DU");
    m_words.push_back("D''");
    m_words.push_back("LE");
    m_words.push_back("LA");
    m_words.push_back(" LES ");
    m_words.push_back("L''");
    m_words.push_back("AU");
    m_words.push_back("AUX");
    m_words.push_back("UN");
    m_words.push_back("UNE");
    m_words.push_back("EN");
    m_words.push_back("ET");
    m_words.push_back("PAR");
    m_words.push_back("SUR");
    m_words.push_back(":");
    m_words.push_back("LECTURE");
    m_words.push_back("MAILLAGE");
    m_words.push_back("POINTS");
    m_words.push_back("NOM");
    m_words.push_back("TYPE");
    m_words.push_back("GEOMETRIQUE");
    m_words.push_back("GLOBALE");
    m_words.push_back("ESPACE");
    
    // load file in string ignoring comment lines and meaningless words
    stringstream sf;
    ifstream ff(filename);
    string line;
    bool intitle=false;
    string b="";

    while (!ff.eof()) {
        getline(ff,line);
        remove_char(line,'\r'); // erase "\r" dos convention
        //remove_leading_blank(line); // remove leading blank
        if (line[0]!='*') {
            if (line.find("TITRE")!=string::npos) {
                intitle=true;
            }

            line=b+line; // add carriage return in first position

            if (line.find("FORMAT")!=string::npos) {
                intitle=false;
            }

            if (!intitle) {
                line=line+"\n"; // add artificial "\n"
                remove_meaningless_words(line,m_words);
                line.erase(line.size()-1,1); // erase last "\n"
            }

            sf << line;
            b="\n";
        }
    }

    ff.close();

    ofstream out("out0.dat");
    out << sf.str();
    out.close();
    
    // file (string) analysis
    string w;
    // TITRE analysis
    sf >> w;
    uint nl=0;
    title="";
    char Li[250];
    title="";

    if (w=="TITRE") {
        sf >> nl;
        sf.getline(Li,250);
        string b="";

        for (uint i=0; i<nl; i++) {
            sf.getline(Li,250);
            title=title+b+string(Li);
            b="\n";
        }
    }
    
    // FORMAT analysis
    sf >> w;
    string fcor="*",fnum="*";

    if (w=="FORMAT") {
        while (w!="DESCRIPTION" && !sf.eof()) {
            sf >> w;

            if (w=="COORDONNEES") {
                sf >> fcor;
                w=fcor;
            }

            if (w=="NUMEROTATION") {
                sf >> fnum;
                w=fnum;
            }
        }
    }

    replace(fcor,"'","");
    replace(fcor,"(","");
    replace(fcor,")","");
    uint n_fcor=0,w_fcor=0,d_fcor=0;
    char T_fcor;

    if (fcor!="*") {
        read_f77_format(fcor,n_fcor,T_fcor,w_fcor,d_fcor);
    }

    replace(fnum,"'","");
    replace(fnum,"(","");
    replace(fnum,")","");
    uint n_fnum=0,w_fnum=0,d_fnum=0;
    char T_fnum;

    if (fnum!="*") {
        read_f77_format(fnum,n_fnum,T_fnum,w_fnum,d_fnum);
    }

    if (w!="DESCRIPTION") {
        message_event("err_load_mel1",msg_data(filename));
    }
    
    // DESCRIPTION analysis
    vector<string> varname;
    uint nbelt=0;

    while (w!="BLOC") {
        sf >> w;

        if (w=="VARIABLES") {
            sf >> w;

            while (w.substr(0,1)=="'") {
                varname.push_back(w);
                sf >> w;
            }
        }

        if (w=="ELEMENTS") {
            sf >> nbelt;
        }
    }

    dim=varname.size(); // mesh dim
    xv=dim>=1;
    yv=dim>=2;
    zv=dim>=3;

    if (w!="BLOC") {
        message_event("err_load_mel2",msg_data(filename));
    }
    
    // BLOC analysis
    map<string,string> elt_geo;
    bool gauss_lobatto=false;
    elt_geo["SEGMENTS"]="SE";
    elt_geo["TRIANGLES"]="TR";
    elt_geo["QUADRANGLES"]="QU";
    elt_geo["TETRAEDRES"]="TE";
    elt_geo["HEXAEDRES"]="HE";
    elt_geo["PRISMES"]="PR";
    map<string,uint> elt_type;
    elt_type["SE01"]=1;
    elt_type["TR01"]=2;
    elt_type["QU01"]=3;
    elt_type["TE01"]=4;
    elt_type["HE01"]=5;
    elt_type["PR01"]=6;
    elt_type["PY01"]=7;
    elt_type["SE02"]=8;
    elt_type["TR02"]=9;
    elt_type["QU02"]=10;
    elt_type["TE02"]=11;
    elt_type["HE02"]=12;
    elt_type["PR02"]=13;
    elt_type["PY02"]=14;
    elt_type["TR03"]=21;
    elt_type["TR04"]=23;
    elt_type["TR05"]=25;
    elt_type["SE03"]=26;
    elt_type["SE04"]=27;
    elt_type["SE05"]=28;
    elt_type["TE03"]=29;
    elt_type["TE04"]=30;
    map<uint,uint> blocs; // description of a block of elements (msh type,number of elements)
    uint nb_elt_bloc=0; // nb of element in current block
    uint ord=0; // order of the elements
    uint nbetot=0;
    bool new_bloc=true;
    bool forme1=false;
    bool forme2=false; // usefull for the next algorithm

    while (new_bloc) { // loop on block
        forme1=false;
        forme2=false;
        sf >> w;
        map<string,string>::iterator it1=elt_geo.find(w);

        if (it1!=elt_geo.end()) { //block type geo found
            sf >> w >> w; //word LAGRANGE ignored

            if (w=="GAUSS-LOBATTO") {
                gauss_lobatto=true;
                sf >> w;
            }

            if (w=="ORDRE") {
                sf >> ord;
            }

            if (w=="DEGRE") {
                sf >> ord;
            }

            if (w[0]=='P' || w[0]=='Q') {
                w.erase(0,1);
                ord=str2int(w);
            }

            if (ord>2) {
                message_event("err_load_mel3");
            }

            string t=it1->second+"0"+ord;
            sf >> nb_elt_bloc;
            blocs[elt_type[t]]=nb_elt_bloc;
            nbetot+=nb_elt_bloc;
            forme1=true;

        } else {
            map<string,uint>::iterator it2=elt_type.find(w);

            if (it2!=elt_type.end()) { // block type elt found
                sf >> nb_elt_bloc;
                blocs[it2->second]=nb_elt_bloc;
                nbetot+=nb_elt_bloc;
                forme2=true;
            }
        }

        if (!forme1 && !forme2) {
            message_event("err_load_mel4");
        }

        sf >> w;

        if (w=="ELEMENTS" && nbetot==nbelt) {
            new_bloc=false;
        }
    }
    
    // read coordinates and indexes of each element
    string s;
    getline(sf,s); // to jump on the next line
    uint l=0; // element index
    map<uint,uint>::const_iterator it; // bloc iterator

    for (it=blocs.begin(); it!=blocs.end(); ++it) { // loop on blocs of elements
        uint t=it->first;
        uint nbn=element_type_list[t].number_of_nodes;

        for (uint k=0; k<it->second; k++) { // loop on elements
            // read coordinates of the nodes
            vector<node> N(nbn);

            if (fcor=="*") { // free format
                for (uint i=0; i<nbn; i++)
                    for (uint j=1; j<=dim; j++) {
                        sf >> N[i](j);
                    }

            } else { // formatted format for real number
                // waiting for nbn*ndim reals (grouped by n_fcor real of length w_fcor)
                uint i=0;
                uint j=1;
                uint nbl=nbn*dim/n_fcor;

                if (nbn*dim % n_fcor>0) {
                    nbl++;
                }

                for (uint li=0; li<nbl; li++) {
                    getline(sf,s);

                    while (s!="") {
                        string sr=s.substr(0,w_fcor);
                        double v=str2float(sr);
                        s.erase(0,w_fcor);
                        N[i](j)=v;
                        j++;

                        if (j==dim+1) {
                            j=1;
                            i++;
                        }
                    }
                }
            }

            // read indexes of nodes and add nodes to the list of nodes
            vector<uint> ind(nbn);

            if (fnum=="*") { // free format
                for (uint i=0; i<nbn; i++) {
                    sf >> ind[i];

                    if (nodes_list.not_in_list(ind[i])) {
                        nodes_list.add(ind[i],N[i]);
                    }
                }

            } else { // formatted format for integers, waiting nbn integers of max length w_fnum
                uint i=0;
                uint nbl=nbn/n_fnum;

                if (nbn% n_fnum>0) {
                    nbl++;
                }

                for (uint li=0; li<nbl; li++) {
                    getline(sf,s);

                    while (s!="") {
                        string sr=s.substr(0,w_fnum);
                        ind[i]=str2int(sr);
                        s.erase(0,w_fnum);

                        if (nodes_list.not_in_list(ind[i])) {
                            nodes_list.add(ind[i],N[i]);
                        }

                        i++;
                    }
                }
            }

            // add element to the list of elements
            l++;
            elements_list.add(l,element(l,t,ind));
        }
    }
    
    // read definition of domain
    sf >> w;
    uint ie,ia;
    enum type_domain {exp_point=0,exp_line,exp_edge,exp_element,imp_element,undef};
    type_domain td=undef;
    uint nud=0;

    while (w!="FIN") {
        if (w=="DOMAINE") { // new domain
            td=undef;
        }

        nud++;
        string dn;
        sf >> dn; // read domain name
        replace(dn,"'","");
        physical_domain pd(nud,dn,this);
        sf >> w >> ie; // read E i or ELEMENT i
        sf >> w; // may be i or / or F(ACE) or A(RETE) or P(OINT) or DOMAIN or END

        if (w[0]=='/') {
            td=imp_element;
        }

        if (w[0]=='F') {
            td=exp_edge;
        }

        if (w[0]=='A') {
            td=exp_line;
        }

        if (w[0]=='P') {
            td=exp_point;
        }

        if (w[0]<='9' && w[0]>'0') {
            td=exp_element;
        }

        switch (td) {
            case imp_element : {
                    pd.dim=dim;
                    sf >> ia;
                    pd.elements.resize(ia-ie+1);

                    for (uint j=ie; j<=ia; j++) {
                        pd.elements[j-ie]=j;
                        elements_list(j).num_domain=nud;
                    }

                    sf >> w;
                    break;
                }
            case exp_element : {
                    pd.dim=dim;
                    pd.elements.push_back(ie);
                    elements_list(ie).num_domain=nud;

                    while (w!="DOMAINE" && w!="FIN") {
                        ie=str2int(w);
                        pd.elements.push_back(ie);
                        elements_list(ie).num_domain=nud;
                        sf >> w;
                    }

                    break;
                }
            case exp_edge :
            case exp_line :
            case exp_point : {
                    if (debug) {
                    	ofdebug << "Domaine " << pd.name << endl;
                    }
                    
                    pd.dim=static_cast<uint>(td);
                    sf >> ia >> w;
                    vector<uint> * E=&element_type_list[elements_list(ie).type].edges[ia-1];
                    vector<uint> nodes_edge((*E).size()-1);

                    for (uint k=1; k<(*E).size(); k++) {
                        nodes_edge[k-1]=elements_list(ie).nodes[(*E)[k]-1];
                    }

                    l++;
                    element elt(l,(*E)[0],nodes_edge,nud);
                    elt.boundary_element.push_back(pair<uint,uint>(ie,ia));
                    if (debug) {
                    	ofdebug << elements_list(ie) << " ia=" << ia << "  ";
                    	ofdebug << (*E)[1] << " "<< (*E)[2] << " " << (*E)[3] << elt << endl;
                    }
                    
                    elements_list.add(l,elt);
                    pd.elements.push_back(l);

                    while (w[0]=='E') {
                        sf >> ie >> w >> ia >> w;
                        E=&element_type_list[elements_list(ie).type].edges[ia-1];
                        nodes_edge.resize((*E).size()-1);

                        for (uint k=1; k<(*E).size(); k++) {
                            nodes_edge[k-1]=elements_list(ie).nodes[(*E)[k]-1];
                        }

                        l++;
                        element elt(l,(*E)[0],nodes_edge,nud);
                        elt.boundary_element.push_back(pair<uint,uint>(ie,ia));
                        
                    	if (debug) {
                    		ofdebug << elements_list(ie) << " ia=" << ia << "  ";
                    		ofdebug << (*E)[1] << " "<< (*E)[2] << " " << (*E)[3] << elt << endl;
                    	}

                        elements_list.add(l,elt);
                        pd.elements.push_back(l);
                    }

                    break;
                }
            case undef :
                message_event("err_load_mel5",msg_data(dn));
        }

        domains_list[nud]=pd;
    }
    
    // end reading of a mel file
    pop_context();
}
