// ____________________________________________________________________________
// classes used to manage a mesh :
//   node             : 3D point
//   list_of_nodes    : list of nodes
//   element          : finite element (geometric data)
//   list_of_elements : list of element
//   element_type     : data associated to a type of element
//   physical_domain  : definition of a physical domain (set of elements)
//   mesh             : data associated to a mesh (list of nodes, list of elements, ...)
//_____________________________________________________________________________
//   mesh
//     |
//     | - list_of_nodes = map<uint,node>
//     |                             |
//     |                             | - ref
//     |                             | - x[3]
//     |
//     |
//     | - list_of_elements = map<uint,element> 
//     |                                  |
//     |         | ---------------------- |
//     |         |
//     |         | - index                        | index, dim, order
//     |         | - type  --------> element_type | names (melina, msh,...)
//     |         | - vector<nodes>                | nb of nodes, vertices, edges
//     |         | - domain num                   | edge numbering
//     |         | - boundary_element
//     |         |      (parent,edge index)
//     |
//     |
//     | - map<uint,physical_domain>
//     |                   |
//     |                   | index, name, dim
//     |                   | vector<uint> (element list)
//     |                   | mesh *
//     |                   | parent domain
//_____________________________________________________________________________
//
// main functionnalities
//
// mesh::loadfromfile -->  loadfrom_msh loadfrom_mel
// mesh::saveto_msh saveto_mel saveto_vtk saveto_amdba
//_____________________________________________________________________________

#ifndef MESH_H_INCLUDED
#define MESH_H_INCLUDED

#include <iostream>
#include <map>
#include <vector>
#include <string>
#include <list>
#include <fstream>
#include <sstream>
#include <algorithm>
#include "message.h"
#include "utils.h"

using namespace std;

typedef unsigned int uint;
enum mesh_type {msh,mel,vtk,amdba,generated};
class mesh;

//=============================================================================
//                       c l a s s    n o d e
//=============================================================================
// point in 3D
class node {
    public :
        double *x; // components
        uint ref; // reference number used by amdba_format
        node(double x1=0, double x2=0, double x3=0, uint r=0) {
            x=new double[3];
            x[0]=x1;
            x[1]=x2;
            x[2]=x3;
            ref=r;
        }
        ~node() {
            delete [] x;
        }
        node(double* X, uint r=0) {
            x[0]=X[0];
            x[1]=X[1];
            x[2]=X[2];
            ref=r;
        }
        node(const node & Y) {
            x=new double[3];
            x[0]=Y.x[0];
            x[1]=Y.x[1];
            x[2]=Y.x[2];
            ref=Y.ref;
        }
        node& operator=(const node & Y) {
            x[0]=Y.x[0];
            x[1]=Y.x[1];
            x[2]=Y.x[2];
            ref=Y.ref;
            return *this;
        }
        vector<double> tovect(int n=3) {
            vector<double> v(n);

            for (int i=0; i<n; i++) {
                v[i]=x[i];
            }

            return v;
        }
        const double & operator()(uint i) const {
            return x[i-1];
        }
        double & operator()(uint i) {
            return x[i-1];
        }
        friend ostream & operator<<(ostream & o, const node & X) {
            o<<"("<<X.x[0]<<","<<X.x[1]<<","<<X.x[2]<<")";
            return o;
        }
        node & translate(const vector<double> & t);
        node & scale(double a, const vector<double> & o);
        node & dilate(const vector<double> & lambda, const vector<double> & o);
        node & rotateXY(double t, const vector<double> & o);
};

typedef map<uint,node>::iterator it_ln;
typedef map<uint,node>::const_iterator cit_ln;

//=============================================================================
//                       c l a s s    l i s t _ o f _ n o d e s
//=============================================================================
// "ordered" list of node, using the stl class map
class list_of_nodes {
    public :
        map<uint,node> nodes; // list of nodes
        void add(uint i, const node & X) {
            nodes.insert(pair<uint,node>(i,X));
        }
        void add(uint i, double x1, double x2, double x3=0.) {
            add(i, node(x1,x2,x3));
        }
        void erase(uint i);
        void clear();
        uint size() const {
            return nodes.size();
        }
        cit_ln begin() const {
            return nodes.begin();
        }
        cit_ln end() const {
            return nodes.end();
        }
        it_ln begin() {
            return nodes.begin();
        }
        it_ln end() {
            return nodes.end();
        }
        node & find(uint i);
        bool not_in_list(uint i);
        node & operator()(uint i);
        uint last_node() const;
        friend ostream & operator<<(ostream & o, const list_of_nodes & L);
};

//=============================================================================
//                       c l a s s    p h y s i c a l _ d o m a i n
//=============================================================================
// data of a physical domain

class physical_domain {
    public :
        mesh * pmesh; // pointer to the mesh
        uint num; // index of the domain
        string name; // readable name
        bool multidim; // true all the elements have not the same dimension
        uint dim; // domain dimension
        vector<uint> elements; // list of all index of the elmnts of the domain
        vector<physical_domain *> parents; // list of parents domain
        
        physical_domain(uint nu=0, const string & na=string(""), mesh * pm=NULL, uint d=0) {
            num=nu;
            name=na;
            multidim=false;
            dim=d;
            pmesh=pm;
        }
        friend ostream & operator<<(ostream &, const physical_domain &);
};

typedef map<uint,physical_domain>::iterator it_ld;
typedef map<uint,physical_domain>::const_iterator cit_ld;

//=============================================================================
//                       c l a s s    d o m _ i n f o
//=============================================================================
// utility class for saving domains modifications
class dom_info {
    public :
        physical_domain * domain;
        bool keep;
        string new_name;
        bool new_def;
        physical_domain * parent_domain;
        bool crack; // true for a crack sub domain
        bool cracked; // true for a domain which is cracked
        
        dom_info(physical_domain * q=NULL, bool k=true, const string & s=string(""), bool nd=false, physical_domain * pd=NULL, bool cr=false, bool crd=false) : domain(q), new_name(s), parent_domain(pd) {
            keep=k;
            new_def=nd;
            crack=cr;
            cracked=crd;
            
        }
        friend ostream & operator<<(ostream & o, const dom_info & di);
};

//=============================================================================
//                       c l a s s    e l e m e n t
//=============================================================================
// geometric data of a finite element

class element {
    public :
        uint num; // index of the element
        uint num_order; // other index
        uint type; // index of the type of the element
        vector<uint> nodes; // list of the indexes of the nodes
        uint num_domain; // index of the physical domain whom element belongs
        uint num_part; // index of the partition whom element belongs
        uint num_geom; // index of the geometry whom element belongs
        list<pair<uint,uint> > boundary_element; // if exists, gives the index of the parents element E and the index of the boundary of E (1,2,3 ..)
        element() {
            num=0;
            type=0;
            num_domain=0;
            num_part=0;
            num_geom=0;
            num_order=0;
        }
        element(uint n, uint t, const vector<uint> & ns, uint nd=0, uint ng=0, uint np=0);
        vector<element> edges() const; // create a list of the edges of the element
        string string_index(); // string index of an elmnt ("node1_node2_...")
        bool is_adjacent_to(const element &); // test if an element has a boundary with an other element
        friend ostream & operator<<(ostream & o, const element & E);
};

// aliases
typedef map<uint,element>::iterator it_le;
typedef map<uint,element>::const_iterator cit_le;

// utility
pair<uint,uint> common_edge(const element& elt1, const element& elt2);

//=============================================================================
//                       c l a s s    l i s t _ o f _ e l e m e n t s
//=============================================================================
// "ordered" list of elements, using the stl class map
class list_of_elements {
    public :
        map<uint,element> elements;  //list of elements
        void add(uint i, const element & E) {
            elements.insert(pair<uint,element>(i,E));
        }
        void erase(uint i);
        void clear();
        uint size() const {
            return elements.size();
        }
        cit_le begin() const {
            return elements.begin();
        }
        cit_le end() const {
            return elements.end();
        }
        it_le begin() {
            return elements.begin();
        }
        it_le end() {
            return elements.end();
        }
        element & find(uint i);
        const element & find(uint i) const;
        element & operator()(uint i);
        const element & operator()(uint i) const;
        uint last_elt() const;
        friend ostream & operator<<(ostream & o, const list_of_elements & L);
};

//=============================================================================
//                       c l a s s    e l e m e n t _t y p e
//=============================================================================
// data associated to a type of element
class element_type {
    public :
        uint index; // index of the element (msh classification)
        uint dim; // dimension of the element
        uint order; // order of the (finite) element
        const char * name; // explicit name (msh)
        const char * mel_name; // explicit name (mel)
        const char * amdba_name; // explicit name (amdba)
        uint vtk_index; // vtk index
        uint number_of_nodes; // number of nodes
        uint number_of_vertices; // number of vertices
        uint number_of_edges; // number of edges
        vector<vector<uint> > edges; // for each edge, type and nodes number (local)
        // msh uses standard numbering (main vertices then "interior" nodes)
        element_type() : name(NULL), mel_name(NULL), amdba_name(NULL) {
            index=0;
            dim=0;
            order=0;
            vtk_index=0;
            number_of_nodes=0;
            number_of_vertices=0;
            number_of_edges=0;
        }
        element_type(uint i, uint d, uint o, const char * n, const char * mn, const char * an, uint vi, uint nn, uint nv, uint ne) : name(n), mel_name(mn), amdba_name(an) {
            index=i;
            dim=d;
            order=o;
            vtk_index=vi;
            number_of_nodes=nn;
            number_of_vertices=nv;
            number_of_edges=ne;
        }
        friend ostream & operator<<(ostream & o, const element_type & E) {
            o<<"\n"<<"Element type "<<E.index<<" : "<<"dim="<<E.dim<<" ordre="<<E.order;
            o<<" nom="<<E.name<<"mel name="<<E.mel_name<<"amdba_name="<<E.amdba_name<<" vtk index="<<E.vtk_index;
            o<<" nb_noeuds="<<E.number_of_nodes<<" nb_sommets="<<E.number_of_vertices;
            o<<" nb_faces="<<E.number_of_edges<<" ";

            for (uint i=0; i<E.number_of_edges; i++) {
                o<<"bord "<<i+1<<" : ";
                for (uint j=0; j<E.edges[i].size(); j++) {
                    o<<E.edges[i][j]<<" ";
                }
            }

            return o;
        }
};

//=============================================================================
//                       c l a s s    m e s h
//=============================================================================
// data associated to a mesh
class mesh {
    public :
        uint dim; // dimension of the mesh
        mesh_type loadfrom; // original file format
        string name; // mesh name (filename)
        string title; // title of the mesh
        double xmin,xmax,ymin,ymax,zmin,zmax; // box cooordinates
        bool xv,yv,zv; // true if (x resp. y, z) not always 0
        list_of_nodes nodes_list; // collection of nodes
        list_of_elements elements_list; // collection of elements
        map<string,element> edges_list; // collection of all edges, with a string index
        map<uint,physical_domain> domains_list; // collection of physical domains
        mesh() {
            dim=0;
            xmin=xmax=ymin=ymax=zmin=zmax=0.;
        }
        mesh(const char * filename) {
            loadfromfile(filename);
        }
        mesh(const mesh & M); // copy constructor
        mesh(double x1, double x2, uint nbx, string & dom_name, vector<string> & boundary_name); // constructor with a regular mesh of a segment
        mesh(double x1, double x2, uint nbx, double y1, double y2, uint nby, string& dom_name, vector<string> & boundary_name); // constructor from a regular mesh of a 2D box
        mesh & operator=(const mesh & M); // equal operator
        mesh & translate(const vector<double> & t); // translate a mesh
        mesh & scale(double alpha, const vector<double> & o=vector<double>(3,0)); // scale a mesh (isotrope)
                     
        mesh & dilate(const vector<double> & lambda, const vector<double> & o=vector<double>(3,0)); // dilate a mesh
        mesh & rotateXY(double theta, const vector<double> & o); // rotate in XY a mesh
        void segment_mesh(uint nbx, string & dom_name, vector<string> &); // mesh a unit segment
                          
        void square_mesh(uint nbx, uint nby, string & dom_name, vector<string> &); // mesh a unit square
                         
// void crown_mesh(uint nbs,uint nbl,double sl,
//                 string& dom_name=string(""),            //mesh a unit crown
//                 vector<string>& boundary_name=vector<string>(2,string("")));
        
        physical_domain * find_domain(const string & na); // find a domain by its name
        uint last_dom() const;
        uint last_node() const {
            return nodes_list.last_node();
        }
        uint last_elt() const {
            return elements_list.last_elt();
        }
        mesh& operator+=(const mesh &); // merge 2 meshes
        void null_coordinates();
        void loadfromfile(const char * filename); // load mesh from file
        void loadfrom_msh(const char * filename); // msh format (gmsh)
        void loadfrom_mel(const char * filename); // mel format (melina)
        void loadfrom_vtk(const char * filename); // vtk format (paraview)
        void loadfrom_amdba(const char * filename); // amdba format(emc2)
        void saveto_msh(const char * filename=NULL); // save mesh to file .msh
        void saveto_mel(const char * filename=NULL, bool silent=true); // save mesh to file .mel
        void saveto_vtk(const char * filename=NULL, int c=-1); // save mesh to file .vtk
        void saveto_amdba(const char * filename=NULL); // save mesh to file .amdba
        bool valid_savefile(string & filename, const string & ext, bool silent=true);
        void print(ostream & o, bool d) const; // print mesh data
        mesh& merge(const mesh & M); // merge mesh M (overlapping are not managed!)
        friend ostream & operator<<(ostream & o, const mesh & M) {
            M.print(o,true);
            return o;
        }
        friend mesh operator+(const mesh& M1,const mesh& M2); // merge two meshes in a new one
        void make_edges_list(); // make the list of all edges
        void crack_sub_domain(physical_domain & pd); // transform a subdomain into a crack
};

//=============================================================================
//                      f e w   f u n c t i o n s
//=============================================================================
void init();
void read_msh_nodes(fstream & ff, mesh & M, const string & v);
void read_msh_elements(fstream & ff, mesh & M, const string & v);
string read_msh_format(fstream & ff, const string & na);
void read_msh_domains(fstream & ff, mesh & M, const string & v);
void complete_msh_data(mesh & M, const string & v);

#endif // MESH_H_INCLUDED
