// ============================================================================
//                  ( g ) m s h   m e s h   u t i l i t i e s
// ============================================================================

#include "mesh.h"
#include <algorithm>
#include <cmath>

extern map<uint,element_type> element_type_list;
extern map<string,mesh_type> mesh_type_list;


// ============================================================================
//                               format MSH 2.0 (gmsh)
// ============================================================================
// $MeshFormat
// version-number file-type data-size
// $EndMeshFormat
// $Nodes
// number-of-nodes
// node-number x-coord y-coord z-coord
// ...
// $EndNodes
// $Elements
// number-of-elements
// elm-number elm-type number-of-tags < tag > ... node-number-list
// ...
// $EndElements
// $PhysicalNames
// number-of-physical-names
// physical-number "physical-name"
// ...
// $EndPhysicalNames
// $NodeData
// number-of-string-tags
// < "string-tag" >
// ...
// number-of-real-tags
// < real-tag >
// ...
// number-of-integer-tags
// < integer-tag >
// ...
// node-number value ...
// ...
// $EndNodeData
// $ElementData
// number-of-string-tags
// < "string-tag" >
// ...
// number-of-real-tags
// < real-tag >
// ...
// number-of-integer-tags
// < integer-tag >
// ...
// elm-number value ...
// ...
// $EndElementData
// $ElementNodeData
// number-of-string-tags
// < "string-tag" >
// ...
// number-of-real-tags
// < real-tag >
// ...
// number-of-integer-tags
// < integer-tag >
// ...
// elm-number number-of-nodes-per-element value ...
// ...
// $ElementEndNodeData
// ============================================================================
//                               format MSH 2.1 (gmsh)
// ============================================================================
// la seule evolution est la suivante :
// $PhysicalNames
// number-of-physical-names
// physical-dimension physical-number "physical-name"
// ...
// $EndPhysicalNames
// ============================================================================
//                               format MSH 2.2 (gmsh)
// ============================================================================
// a venir

// msh mesh loading
void mesh::loadfrom_msh(const char * filename) {
    push_context("mesh::loadfrom_msh");
    string kw;
    bool cont=true;
    string v;
    map<string,int> keys;
    keys["$MeshFormat"]=1;
    keys["$Nodes"]=2;
    keys["$PhysicalNames"]=3;
    keys["$Elements"]=4;
    keys["$ElementData"]=5; // ignored
    keys["$NodeData"]=6; // ignored
    keys["$ElementNodeData"]=7; // ignored
    fstream ff(filename,fstream::in);

    // main loop
    v="2";
    while (cont && !ff.eof()) {
        ff >> kw;

        if (keys.find(kw)!=keys.end()) {
            switch (keys[kw]) {
                case 1:
                    v=read_msh_format(ff,string(filename));
                    kw="";
                    break;
                case 2:
                    read_msh_nodes(ff,*this,v);
                    kw="";
                    break;
                case 3:
                    read_msh_domains(ff,*this,v);
                    kw="";
                    break;
                case 4:
                    read_msh_elements(ff,*this,v);
                    kw="";
                    break;
                default :
                    kw="";
                    break;
            }
        }
    }

    ff.close();
	// data completion
    complete_msh_data(*this,v);
    pop_context();
}

// read the file format
string read_msh_format(fstream & ff, const string & na) {
    push_context("read_msh_format");
    uint i1,l;
    string v,s;
    ff >> v >> i1 >> l >> s;
    
    if (v.find("2")!=0) {
        msg_data md(na);
        md << "msh 2.x";
        message_event("err_format",md);
    }
    
    pop_context();
    return v;
}

// read the definition of physical domains
void read_msh_domains(fstream & ff, mesh & M, const string & v) {
    push_context("read_msh_domains");
    uint nbpname,pnum,pdim;
    string pname,s;
    map<string,int> vs;//=vector<string>("2","2.1","2.2");
    
    vs["2"]=0;
    vs["2.1"]=1;
    vs["2.2"]=2;

    ff>>nbpname;
    
    for (uint i=0; i<nbpname; i++) {
        switch (vs[v]) {
            case 0 :
                ff >> pnum >> pname;
                break;
            case 1 :
                ff >> pdim >> pnum >> pname;
                break;
            default :
                ff >> pdim >> pnum >> pname;
                break;
        }
        

        if (pname=="\"\"") {
            pname="\""+int2str(pnum)+"\"";
        }

        M.domains_list[pnum]=physical_domain(pnum,pname,&M);
    }

    ff >> s;
    pop_context();
}

// read the number of nodes and nodes (node-number x-coord y-coord z-coord)
void read_msh_nodes(fstream & ff, mesh & M, const string & v) {
    push_context("read_msh_nodes");
    uint nn,num;
    string s;
    double x,y,z;
    ff>>nn;

    for (uint i=0; i<nn; i++) {
        ff >> num >> x >> y >> z;
        //cout<<num<<" "<<node(x,y,z)<<"\n";
        M.xmin=min(M.xmin,x);
        M.xmax=max(M.xmax,x);
        M.ymin=min(M.ymin,y);
        M.ymax=max(M.ymax,y);
        M.zmin=min(M.zmin,z);
        M.zmax=max(M.zmax,z);
        M.nodes_list.add(num,node(x,y,z));
    }

    ff >> s;
    pop_context();
}

// read the elements (elm-number elm-type number-of-tags < tag > ... node-number-list)
void read_msh_elements(fstream & ff, mesh & M, const string & v) {
    push_context("read_msh_elements");
    uint elts_number,num,typ,ntag,nbn,i1,nd,ng,np;
    string s;
    ff >> elts_number;

    for (uint i=0; i<elts_number; i++) {
        nd=0,ng=0,np=0;
        ff >> num >> typ >> ntag;

        if (ntag > 0) {
            uint *tags=new uint[ntag];

            for (uint j=0; j<ntag; j++) {
                ff >> tags[j];
            }

            nd=tags[0];

            if (ntag>1) {
                ng=tags[1];
            }

            if (ntag>2) {
                np=tags[2];
            }

            delete [] tags;
        }

        nbn=element_type_list[typ].number_of_nodes;
        vector<uint> nodes;

        for (uint j=0; j<nbn; j++) {
            ff >> i1;
            nodes.push_back(i1);
        }

        M.elements_list.add(num,element(num,typ,nodes,nd,ng,np));
    }

    ff >> s;
    pop_context();
}

// data completion after loading a msh file
// find parent element for boundary elements
// find mesh dimension and physical domain dimension
// make list of elements of a physical domain

void complete_msh_data(mesh & M, const string & v) {
    push_context("complete_msh_data");
    
    // physical domains analysis
    // make lists of elements of the physical domains
    uint d;
    M.dim=0;

    for (cit_le it=M.elements_list.begin(); it!=M.elements_list.end(); ++it) {
        const element & E=it->second;
        d=E.num_domain;

        if (M.domains_list.find(d)==M.domains_list.end()) { // new domain
            M.domains_list[d]=physical_domain(d,string("domain_")+d,&M);
        }

        physical_domain & Dom=M.domains_list[d];
        Dom.elements.push_back(E.num);

        // find dimension of mesh and domains
        if (Dom.dim==0) {
            Dom.dim=element_type_list[E.type].dim;
        }

        if (Dom.dim!=0 && element_type_list[E.type].dim>Dom.dim) {
            Dom.dim=element_type_list[E.type].dim;
            Dom.multidim=true;
        }

        M.dim=max(M.dim,element_type_list[E.type].dim);
    }

	// split multidimensional domains in monodimensionnal domains
	// original domain is kept (EXPERIMENTAL)
    map<uint,physical_domain>::iterator it;
    uint last_num=0;

    for (it=M.domains_list.begin(); it!=M.domains_list.end(); ++it) {
        last_num=max(last_num,it->first);
    }

    for (it=M.domains_list.begin(); it!=M.domains_list.end(); ++it) {
        physical_domain & Dom=it->second;

        if (Dom.multidim) {
            uint dims[4]= {0,0,0,0}; // find multiple dimensions

            for (uint i=0; i<Dom.elements.size(); i++) {
                const element & E=M.elements_list.elements[Dom.elements[i]];
                uint t=E.type;
                dims[element_type_list[t].dim]=1;
            }

            for (uint k=0; k<4; k++) { // create new domain
                if (dims[k]!=0) {
                    last_num++;
                    string na=Dom.name+string("_")+dims[k];
                    physical_domain p(last_num,na,&M,dims[k]);
                    p.parents.push_back(&Dom);
                    M.domains_list[last_num]=p;
                    dims[k]=last_num;
                }
            }

            for (uint i=0; i<Dom.elements.size(); i++) { // fill new domain
                const element & E=M.elements_list.elements[Dom.elements[i]];
                uint d=element_type_list[E.type].dim;
                physical_domain & Dom=M.domains_list[dims[d]];
                Dom.elements.push_back(Dom.elements[i]);
            }
        }
    }
    
    // build parent relations of elements belonging to a subdomain (dim < mesh.dim)
    // !!!! only for n-1 dimension subdomain !!!!
    // create boundary elements of all elements of the mesh
    multimap<string,pair<uint,uint> > b_elt;
    uint ind=0;

    for (cit_le it=M.elements_list.begin(); it!=M.elements_list.end(); ++it) {
        const element & E=it->second;
        const element_type & et=element_type_list[E.type];

        if (E.type==15) {
            break; // point element, no boundary
        }

        for (uint i=0; i<et.number_of_edges; i++) { // loop on the edges
            uint m=et.edges[i].size()-1;
            uint t=et.edges[i][0];
            vector<uint> edgi(m);

            for (uint j=1; j<=m; j++) {
                edgi[j-1]=E.nodes[et.edges[i][j]-1];
            }

            sort(edgi.begin(),edgi.end());
            stringstream key;
            key<<t;

            for (uint k=0; k<edgi.size(); k++) {
                key<<"_"<<edgi[k];
            }

            ind++;
            pair<uint,uint> pa=pair<uint,uint>(E.num,i+1);
            b_elt.insert(pair<string,pair<uint,uint> >(key.str(),pa));
        }
    }
    
    // build parent relations for each element of the mesh
    for (it_le it=M.elements_list.begin(); it!=M.elements_list.end(); ++it) {
        element & E=it->second;
        const element_type & et=element_type_list[E.type];
		        
        if (et.dim<M.dim) { // no parent for mesh.dim element !
            vector<uint> no(E.nodes);
            sort(no.begin(),no.end());
            stringstream key;
            key<<E.type;

            for (uint k=0; k<no.size(); k++) {
                key<<"_"<<no[k];
            }

            typedef multimap<string,pair<uint,uint> >::const_iterator itm;
            pair<itm,itm> pit=b_elt.equal_range(key.str());

            for (itm jt=pit.first; jt!=pit.second; ++jt) {
                E.boundary_element.push_back(jt->second); // a boundary element is found
            }
        }
    }
    
    // analysis of the subdomains : what are the physical parent domains (boundary of), 3 cases :
    //    no parent  : abnormal situation (parent does not belong to any physical domain !)
    //    one parent : external boundary of a domain
    //    more than one : internal boundary
    for (it=M.domains_list.begin(); it!=M.domains_list.end(); ++it) {
        physical_domain& Dom=it->second;

        if (Dom.dim<M.dim) { // only subdomains are concerned
            for (uint l=0; l<Dom.elements.size(); l++) {
                element& E=M.elements_list(Dom.elements[l]);
                list<pair<uint,uint> >::const_iterator itb;

                for (itb=E.boundary_element.begin(); itb!=E.boundary_element.end(); ++itb) { // loop on parent elements of element
                    uint nd=M.elements_list(itb->first).num_domain; // domain index of parent (nd=0 -> parents does not belong to any domain !!!)

                    if (nd>0) {
                        uint j=0;
                        bool newp=true;

                        while (j<Dom.parents.size() && newp) {
                            if (nd==Dom.parents[j]->num) {
                                newp=false;
                            }

                            j++;
                        }

                        if (newp) {
                            Dom.parents.push_back(&M.domains_list[nd]);
                        }
                    }
                }
            }
        }
    }
    
    // detect null coordinates
    M.null_coordinates();
    pop_context();
}

// save mesh to msh format file
void mesh::saveto_msh(const char * filename) {
    push_context("mesh::saveto_msh");

    // open file
    string savename;

    if (filename!=NULL) {
        savename=string(filename);
    }

    if (!valid_savefile(savename,string("msh"))) {
        return;
    }

    ofstream f(savename.c_str(),ios::out | ios::binary);

    if (f.fail()) {
        message_event("open_fail",msg_data(savename));
    }
    
    // save msh header
    f << "$MeshFormat\n2 0 8\n$EndMeshFormat";
    // save physical names
    f << "\n$PhysicalNames"<< endl << domains_list.size() << endl;

    for (map<uint,physical_domain>::const_iterator it=domains_list.begin(); it!=domains_list.end(); ++it) {
        f << it->second.num << " " << it->second.name << endl ;
    }

    f << "$EndPhysicalNames" << endl;
    // nodes list
    f << "$Nodes" << endl << nodes_list.size() << endl;

    for (cit_ln it=nodes_list.begin(); it!=nodes_list.end(); ++it) {
        const node & N=it->second;
        f << it->first << " " << N(1) << " " << N(2) << " " << N(3) << endl;
    }

    f << "$EndNodes" << endl;
    // elements list
    f <<"$Elements" << endl << elements_list.size() << endl;

    for (cit_le it=elements_list.begin(); it!=elements_list.end(); ++it) {
        const element & E=it->second;
        f << E.num << " " << E.type ;
        f <<" 3 " << E.num_domain << " " << E.num_geom << " " << E.num_part;

        for (uint i=0; i<E.nodes.size(); i++) {
            f << " " << E.nodes[i];
        }
        f << endl;
    }

    f << "$EndElements" << endl;
    // close file
    f.close();
    pop_context();
}
